@@ -12,30 +12,28 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+# pylint: disable=too-many-lines,too-many-public-methods,too-many-statements,no-self-use
+
+import base64
+import copy
+
 from selenium import webdriver
 
-from .mobilecommand import MobileCommand as Command
-from .errorhandler import MobileErrorHandler
-from .switch_to import MobileSwitchTo
-from .webelement import WebElement as MobileWebElement
-from .imagelement import ImageElement
+from selenium.webdriver.common.by import By
+from selenium.webdriver.support.ui import WebDriverWait
+from selenium.common.exceptions import TimeoutException, InvalidArgumentException
+
+from selenium.webdriver.remote.command import Command as RemoteCommand
 
 from appium.webdriver.clipboard_content_type import ClipboardContentType
 from appium.webdriver.common.mobileby import MobileBy
 from appium.webdriver.common.touch_action import TouchAction
 from appium.webdriver.common.multi_action import MultiAction
 
-from selenium.webdriver.common.by import By
-from selenium.webdriver.support.ui import WebDriverWait
-from selenium.common.exceptions import (TimeoutException,
-        WebDriverException, InvalidArgumentException, NoSuchElementException)
-
-from selenium.webdriver.remote.command import Command as RemoteCommand
-
-import base64
-import copy
-
-DEFAULT_MATCH_THRESHOLD = 0.5
+from .mobilecommand import MobileCommand as Command
+from .errorhandler import MobileErrorHandler
+from .switch_to import MobileSwitchTo
+from .webelement import WebElement as MobileWebElement
 
 # From remote/webdriver.py
 _W3C_CAPABILITY_NAMES = frozenset([
@@ -62,31 +60,33 @@ _FORCE_MJSONWP = 'forceMjsonwp'
 
 # override
 # Add appium prefix for the non-W3C capabilities
+
+
 def _make_w3c_caps(caps):
     appium_prefix = 'appium:'
 
     caps = copy.deepcopy(caps)
     profile = caps.get('firefox_profile')
-    always_match = {}
+    first_match = {}
     if caps.get('proxy') and caps['proxy'].get('proxyType'):
         caps['proxy']['proxyType'] = caps['proxy']['proxyType'].lower()
     for k, v in caps.items():
         if v and k in _OSS_W3C_CONVERSION:
-            always_match[_OSS_W3C_CONVERSION[k]] = v.lower() if k == 'platform' else v
+            first_match[_OSS_W3C_CONVERSION[k]] = v.lower() if k == 'platform' else v
         if k in _W3C_CAPABILITY_NAMES or _EXTENSION_CAPABILITY in k:
-            always_match[k] = v
+            first_match[k] = v
         else:
             if not k.startswith(appium_prefix):
-                always_match[appium_prefix + k] = v
+                first_match[appium_prefix + k] = v
     if profile:
-        moz_opts = always_match.get('moz:firefoxOptions', {})
+        moz_opts = first_match.get('moz:firefoxOptions', {})
         # If it's already present, assume the caller did that intentionally.
         if 'profile' not in moz_opts:
             # Don't mutate the original capabilities.
             new_opts = copy.deepcopy(moz_opts)
             new_opts['profile'] = profile
-            always_match['moz:firefoxOptions'] = new_opts
-    return {'firstMatch': [{}], 'alwaysMatch': always_match}
+            first_match['moz:firefoxOptions'] = new_opts
+    return {'firstMatch': [first_match]}
 
 
 class WebDriver(webdriver.Remote):
@@ -107,8 +107,10 @@ class WebDriver(webdriver.Remote):
         By.IOS_PREDICATE = MobileBy.IOS_PREDICATE
         By.IOS_CLASS_CHAIN = MobileBy.IOS_CLASS_CHAIN
         By.ANDROID_UIAUTOMATOR = MobileBy.ANDROID_UIAUTOMATOR
+        By.ANDROID_VIEWTAG = MobileBy.ANDROID_VIEWTAG
         By.ACCESSIBILITY_ID = MobileBy.ACCESSIBILITY_ID
         By.IMAGE = MobileBy.IMAGE
+        By.CUSTOM = MobileBy.CUSTOM
 
     def start_session(self, capabilities, browser_profile=None):
         """
@@ -201,20 +203,19 @@ class WebDriver(webdriver.Remote):
 
         :rtype: WebElement
         """
+        # TODO: If we need, we should enable below converter for Web context
         # if self.w3c:
-            # if by == By.ID:
-            #     by = By.CSS_SELECTOR
-            #     value = '[id="%s"]' % value
-            # elif by == By.TAG_NAME:
-            #     by = By.CSS_SELECTOR
-            # elif by == By.CLASS_NAME:
-            #     by = By.CSS_SELECTOR
-            #     value = ".%s" % value
-            # elif by == By.NAME:
-            #     by = By.CSS_SELECTOR
-            #     value = '[name="%s"]' % value
-        if by == By.IMAGE:
-            return self.find_element_by_image(value)
+        #     if by == By.ID:
+        #         by = By.CSS_SELECTOR
+        #         value = '[id="%s"]' % value
+        #     elif by == By.TAG_NAME:
+        #         by = By.CSS_SELECTOR
+        #     elif by == By.CLASS_NAME:
+        #         by = By.CSS_SELECTOR
+        #         value = ".%s" % value
+        #     elif by == By.NAME:
+        #         by = By.CSS_SELECTOR
+        #         value = '[name="%s"]' % value
 
         return self.execute(RemoteCommand.FIND_ELEMENT, {
             'using': by,
@@ -230,25 +231,23 @@ class WebDriver(webdriver.Remote):
 
         :rtype: list of WebElement
         """
+        # TODO: If we need, we should enable below converter for Web context
         # if self.w3c:
-            # if by == By.ID:
-            #     by = By.CSS_SELECTOR
-            #     value = '[id="%s"]' % value
-            # elif by == By.TAG_NAME:
-            #     by = By.CSS_SELECTOR
-            # elif by == By.CLASS_NAME:
-            #     by = By.CSS_SELECTOR
-            #     value = ".%s" % value
-            # elif by == By.NAME:
-            #     by = By.CSS_SELECTOR
-            #     value = '[name="%s"]' % value
+        #     if by == By.ID:
+        #         by = By.CSS_SELECTOR
+        #         value = '[id="%s"]' % value
+        #     elif by == By.TAG_NAME:
+        #         by = By.CSS_SELECTOR
+        #     elif by == By.CLASS_NAME:
+        #         by = By.CSS_SELECTOR
+        #         value = ".%s" % value
+        #     elif by == By.NAME:
+        #         by = By.CSS_SELECTOR
+        #         value = '[name="%s"]' % value
 
         # Return empty list if driver returns null
         # See https://github.com/SeleniumHQ/selenium/issues/4555
 
-        if by == By.IMAGE:
-            return self.find_elements_by_image(value)
-
         return self.execute(RemoteCommand.FIND_ELEMENTS, {
             'using': by,
             'value': value})['value'] or []
@@ -341,75 +340,110 @@ class WebDriver(webdriver.Remote):
         """
         return self.find_elements(by=By.ANDROID_UIAUTOMATOR, value=uia_string)
 
-    def find_element_by_image(self, png_img_path,
-                              match_threshold=DEFAULT_MATCH_THRESHOLD):
+    def find_element_by_android_viewtag(self, tag):
+        """Finds element by [View#tags](https://developer.android.com/reference/android/view/View#tags) in Android.
+        It works with [Espresso Driver](https://github.com/appium/appium-espresso-driver).
+
+        :Args:
+         - tag - The tag name of the view to look for
+
+        :Usage:
+            driver.find_element_by_android_viewtag('a tag name')
+        """
+        return self.find_element(by=By.ANDROID_VIEWTAG, value=tag)
+
+    def find_elements_by_android_viewtag(self, tag):
+        """Finds element by [View#tags](https://developer.android.com/reference/android/view/View#tags) in Android.
+        It works with [Espresso Driver](https://github.com/appium/appium-espresso-driver).
+
+        :Args:
+         - tag - The tag name of the view to look for
+
+        :Usage:
+            driver.find_elements_by_android_viewtag('a tag name')
+        """
+        return self.find_elements(by=By.ANDROID_VIEWTAG, value=tag)
+
+    def find_element_by_image(self, img_path):
         """Finds a portion of a screenshot by an image.
         Uses driver.find_image_occurrence under the hood.
 
         :Args:
-        - png_img_path - a string corresponding to the path of a PNG image
-        - match_threshold - a double between 0 and 1 below which matches will
-          be rejected as element not found
+        - img_path - a string corresponding to the path of a image
 
-        :return: an ImageElement object
+        :return: an Element object
         """
-        screenshot = self.get_screenshot_as_base64()
-        with open(png_img_path, 'rb') as png_file:
-            b64_data = base64.b64encode(png_file.read()).decode('UTF-8')
-        try:
-            res = self.find_image_occurrence(screenshot, b64_data,
-                                             threshold=match_threshold)
-        except WebDriverException as e:
-            if 'Cannot find any occurrences' in str(e):
-                raise NoSuchElementException(e)
-            raise
-        rect = res['rect']
-        return ImageElement(self, rect['x'], rect['y'], rect['width'],
-                            rect['height'])
+        with open(img_path, 'rb') as i_file:
+            b64_data = base64.b64encode(i_file.read()).decode('UTF-8')
 
-    def find_elements_by_image(self, png_img_path,
-                               match_threshold=DEFAULT_MATCH_THRESHOLD):
+        return self.find_element(by=By.IMAGE, value=b64_data)
+
+    def find_elements_by_image(self, img_path):
         """Finds a portion of a screenshot by an image.
         Uses driver.find_image_occurrence under the hood. Note that this will
         only ever return at most one element
 
         :Args:
-        - png_img_path - a string corresponding to the path of a PNG image
-        - match_threshold - a double between 0 and 1 below which matches will
-          be rejected as element not found
+        - img_path - a string corresponding to the path of a image
 
-        :return: possibly-empty list of ImageElements
+        :return: possibly-empty list of Elements
         """
-        els = []
-        try:
-            els.append(self.find_element_by_image(png_img_path, match_threshold))
-        except NoSuchElementException:
-            pass
-        return els
+        with open(img_path, 'rb') as i_file:
+            b64_data = base64.b64encode(i_file.read()).decode('UTF-8')
 
-    def find_element_by_accessibility_id(self, id):
+        return self.find_elements(by=By.IMAGE, value=b64_data)
+
+    def find_element_by_accessibility_id(self, accessibility_id):
         """Finds an element by accessibility id.
 
         :Args:
-         - id - a string corresponding to a recursive element search using the
+         - accessibility_id - a string corresponding to a recursive element search using the
          Id/Name that the native Accessibility options utilize
 
         :Usage:
             driver.find_element_by_accessibility_id()
         """
-        return self.find_element(by=By.ACCESSIBILITY_ID, value=id)
+        return self.find_element(by=By.ACCESSIBILITY_ID, value=accessibility_id)
 
-    def find_elements_by_accessibility_id(self, id):
+    def find_elements_by_accessibility_id(self, accessibility_id):
         """Finds elements by accessibility id.
 
         :Args:
-         - id - a string corresponding to a recursive element search using the
+         - accessibility_id - a string corresponding to a recursive element search using the
          Id/Name that the native Accessibility options utilize
 
         :Usage:
             driver.find_elements_by_accessibility_id()
         """
-        return self.find_elements(by=By.ACCESSIBILITY_ID, value=id)
+        return self.find_elements(by=By.ACCESSIBILITY_ID, value=accessibility_id)
+
+    def find_element_by_custom(self, selector):
+        """Finds an element in conjunction with a custom element finding plugin
+
+        :Args:
+         - selector - a string of the form "module:selector", where "module" is
+           the shortcut name given in the customFindModules capability, and
+           "selector" is the string that will be passed to the custom element
+           finding plugin itself
+
+        :Usage:
+            driver.find_element_by_custom("foo:bar")
+        """
+        return self.find_element(by=By.CUSTOM, value=selector)
+
+    def find_elements_by_custom(self, selector):
+        """Finds elements in conjunction with a custom element finding plugin
+
+        :Args:
+         - selector - a string of the form "module:selector", where "module" is
+           the shortcut name given in the customFindModules capability, and
+           "selector" is the string that will be passed to the custom element
+           finding plugin itself
+
+        :Usage:
+            driver.find_elements_by_custom("foo:bar")
+        """
+        return self.find_elements(by=By.CUSTOM, value=selector)
 
     def create_web_element(self, element_id):
         """
@@ -420,18 +454,27 @@ class WebDriver(webdriver.Remote):
         return MobileWebElement(self, element_id)
 
     # convenience method added to Appium (NOT Selenium 3)
-    def scroll(self, origin_el, destination_el):
+    def scroll(self, origin_el, destination_el, duration=None):
         """Scrolls from one element to another
 
         :Args:
          - originalEl - the element from which to being scrolling
          - destinationEl - the element to scroll to
+         - duration - a duration after pressing originalEl and move the element to destinationEl. Default is 600 ms for W3C spec. Zero for MJSONWP.
 
         :Usage:
             driver.scroll(el1, el2)
         """
+
+        # XCUITest x W3C spec has no duration by default in server side
+        if self.w3c and duration is None:
+            duration = 600
+
         action = TouchAction(self)
-        action.press(origin_el).move_to(destination_el).release().perform()
+        if duration is None:
+            action.press(origin_el).move_to(destination_el).release().perform()
+        else:
+            action.press(origin_el).wait(duration).move_to(destination_el).release().perform()
         return self
 
     # convenience method added to Appium (NOT Selenium 3)
@@ -674,6 +717,20 @@ class WebDriver(webdriver.Remote):
         self.execute(Command.LONG_PRESS_KEYCODE, data)
         return self
 
+    def press_button(self, button_name):
+        """Sends a physical button name to the device to simulate the user pressing. iOS only.
+        Possible button names can be found in
+        https://github.com/appium/WebDriverAgent/blob/master/WebDriverAgentLib/Categories/XCUIDevice%2BFBHelpers.h
+
+        :Args:
+         - button_name - the button name to be sent to the device
+        """
+        data = {
+            'name': button_name
+        }
+        self.execute_script('mobile: pressButton', data)
+        return self
+
     @property
     def current_activity(self):
         """Retrieves the current activity running on the device.
@@ -1090,6 +1147,12 @@ class WebDriver(webdriver.Remote):
         self.execute(Command.UPDATE_SETTINGS, data)
         return self
 
+    def toggle_wifi(self):
+        """Toggle the wifi on the device, Android only.
+        """
+        self.execute(Command.TOGGLE_WIFI, {})
+        return self
+
     def toggle_location_services(self):
         """Toggle the location services on the device. Android only.
         """
@@ -1376,6 +1439,16 @@ class WebDriver(webdriver.Remote):
         """
         return self.execute_script('mobile: batteryInfo')
 
+    def finger_print(self, finger_id):
+        """
+        Authenticate users by using their finger print scans on supported emulators.
+
+        :param finger_id: Finger prints stored in Android Keystore system (from 1 to 10)
+        """
+        return self.execute(Command.FINGER_PRINT, {'fingerprintId': finger_id})['value']
+
+    # pylint: disable=protected-access
+
     def _addCommands(self):
         self.command_executor._commands[Command.CONTEXTS] = \
             ('GET', '/session/$sessionId/contexts')
@@ -1472,6 +1545,8 @@ class WebDriver(webdriver.Remote):
             ('POST', '/session/$sessionId/appium/settings')
         self.command_executor._commands[Command.TOGGLE_LOCATION_SERVICES] = \
             ('POST', '/session/$sessionId/appium/device/toggle_location_services')
+        self.command_executor._commands[Command.TOGGLE_WIFI] = \
+            ('POST', '/session/$sessionId/appium/device/toggle_wifi')
         self.command_executor._commands[Command.SET_LOCATION] = \
             ('POST', '/session/$sessionId/location')
         self.command_executor._commands[Command.LOCATION_IN_VIEW] = \
@@ -1490,3 +1565,5 @@ class WebDriver(webdriver.Remote):
             ('POST', '/session/$sessionId/appium/device/get_clipboard')
         self.command_executor._commands[Command.COMPARE_IMAGES] = \
             ('POST', '/session/$sessionId/appium/compare_images')
+        self.command_executor._commands[Command.FINGER_PRINT] = \
+            ('POST', '/session/$sessionId/appium/device/finger_print')
